searchNodes=[{"doc":"Installation Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.7.0&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected. Goals of the library Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No runtime dependencies, to simplify integration with other tools. Background There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest. Sourceror's AST Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropiate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing coments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted. License Copyright (c) 2021 dorgandash@gmail.com Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"Sourceror.html","title":"Sourceror","type":"module"},{"doc":"Appends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#append_comments/3","title":"Sourceror.append_comments/3","type":"function"},{"doc":"Compares two positions. Returns :gt if the first position comes after the second one, and :lt for vice versa. If the two positions are equal, :eq is returned. nil values for lines or columns are coalesced to 0 for integer comparisons.","ref":"Sourceror.html#compare_positions/2","title":"Sourceror.compare_positions/2","type":"function"},{"doc":"Shifts the line numbers of the node or metadata by the given line_correction . This function will update the :line , :closing , :do , :end and :end_of_expression line numbers of the node metadata if such fields are present.","ref":"Sourceror.html#correct_lines/3","title":"Sourceror.correct_lines/3","type":"function"},{"doc":"Returns the arguments of the node. iex&gt; Sourceror . get_args ( { :foo , [ ] , [ { :__block__ , [ ] , [ :ok ] } ] } ) [ { :__block__ , [ ] , [ :ok ] } ]","ref":"Sourceror.html#get_args/1","title":"Sourceror.get_args/1","type":"function"},{"doc":"Returns the column of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the column number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_column ( { :foo , [ column : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_column ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_column/2","title":"Sourceror.get_column/2","type":"function"},{"doc":"Returns the line where the given node ends. It recursively checks for end , closing and end_of_expression line numbers. If none is found, the default value is returned(defaults to 1). iex&gt; Sourceror . get_end_line ( { :foo , [ end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] ] , [ ] } ) 2 iex&gt; Sourceror . get_end_line ( { :foo , [ end_of_expression : [ line : 5 ] ] , [ ] } ) 5 iex&gt; Sourceror . get_end_line ( { :foo , [ closing : [ line : 2 ] , end : [ line : 4 ] ] , [ ] } ) 4 iex&gt; &quot;&quot;&quot; ...&gt; alias Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) |&gt; Sourceror . get_end_line ( ) 3","ref":"Sourceror.html#get_end_line/2","title":"Sourceror.get_end_line/2","type":"function"},{"doc":"Returns the end position of the quoted expression. It recursively checks for end , closing and end_of_expression positions. If none is found, the default value is returned(defaults to [line: 1, column: 1] ). iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; A.{ ...&gt; B ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; foo( ...&gt; :a, ...&gt; :b ...&gt; ) ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_end_position ( quoted ) [ line : 4 , column : 4 ]","ref":"Sourceror.html#get_end_position/2","title":"Sourceror.get_end_position/2","type":"function"},{"doc":"Returns the line of a node. If none is found, the default value is returned(defaults to 1). A default of nil may also be provided if the line number is meant to be coalesced with a value that is not known upfront. iex&gt; Sourceror . get_line ( { :foo , [ line : 5 ] , [ ] } ) 5 iex&gt; Sourceror . get_line ( { :foo , [ ] , [ ] } , 3 ) 3","ref":"Sourceror.html#get_line/2","title":"Sourceror.get_line/2","type":"function"},{"doc":"Returns the metadata of the given node. iex&gt; Sourceror . get_meta ( { :foo , [ line : 5 ] , [ ] } ) [ line : 5 ]","ref":"Sourceror.html#get_meta/1","title":"Sourceror.get_meta/1","type":"function"},{"doc":"Gets the range used by the given quoted expression in the source code. The quoted expression must have at least line and column metadata, otherwise it is not possible to calculate an accurate range, or to calculate it at all. This function is most useful when used after Sourceror.parse_string/1 , before any kind of modification to the AST. The range is a map with :start and :end positions. iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; def foo do ...&gt; :ok ...&gt; end ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 4 ] } iex&gt; quoted = ~S&quot;&quot;&quot; ...&gt; Foo.{ ...&gt; Bar ...&gt; } ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) iex&gt; Sourceror . get_range ( quoted ) %{ start : [ line : 1 , column : 1 ] , end : [ line : 3 , column : 2 ] }","ref":"Sourceror.html#get_range/1","title":"Sourceror.get_range/1","type":"function"},{"doc":"Returns the start position of a node. iex&gt; quoted = Sourceror . parse_string! ( &quot; :foo&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 2 ] iex&gt; quoted = Sourceror . parse_string! ( &quot; \\n \\n foo()&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 3 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;Foo.{Bar}&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo[:bar]&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ] iex&gt; quoted = Sourceror . parse_string! ( &quot;foo(:bar)&quot; ) iex&gt; Sourceror . get_start_position ( quoted ) [ line : 1 , column : 1 ]","ref":"Sourceror.html#get_start_position/2","title":"Sourceror.get_start_position/2","type":"function"},{"doc":"Parses a single expression from the given string. It tries to parse on a per-line basis. Returns {:ok, quoted, rest} on success or {:error, source} on error. Examples iex&gt; ~S&quot;&quot;&quot; ...&gt; 42 ...&gt; ...&gt; :ok ...&gt; &quot;&quot;&quot; |&gt; Sourceror . parse_expression ( ) { :ok , { :__block__ , [ trailing_comments : [ ] , leading_comments : [ ] , token : &quot;42&quot; , line : 2 , column : 1 ] , [ 42 ] } , &quot; \\n :ok&quot; } Options :from_line - The line at where the parsing should start. Defaults to 1 .","ref":"Sourceror.html#parse_expression/2","title":"Sourceror.parse_expression/2","type":"function"},{"doc":"Parses the source code into an extended AST suitable for source manipulation as described in Code.quoted_to_algebra/2 . Two additional fields are added to nodes metadata: :leading_comments - a list holding the comments found before the node. :trailing_comments - a list holding the comments found before the end of the node. For example, comments right before the end keyword. Comments are the same maps returned by Code.string_to_quoted_with_comments/2 .","ref":"Sourceror.html#parse_string/1","title":"Sourceror.parse_string/1","type":"function"},{"doc":"Same as parse_string/1 but raises on error.","ref":"Sourceror.html#parse_string!/1","title":"Sourceror.parse_string!/1","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression. See postwalk/3 for more information.","ref":"Sourceror.html#postwalk/2","title":"Sourceror.postwalk/2","type":"function"},{"doc":"Performs a depth-first post-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.postwalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#postwalk/3","title":"Sourceror.postwalk/3","type":"function"},{"doc":"Prepends comments to the leading or trailing comments of a node.","ref":"Sourceror.html#prepend_comments/3","title":"Sourceror.prepend_comments/3","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression. See prewalk/3 for more information.","ref":"Sourceror.html#prewalk/2","title":"Sourceror.prewalk/2","type":"function"},{"doc":"Performs a depth-first pre-order traversal of a quoted expression with an accumulator. fun is a function that will receive the current node as a first argument and the traversal state as the second one. It must return a {quoted, state} , in the same way it would return {quoted, acc} when using Macro.prewalk/3 . The state is a map with the following keys: :acc - The accumulator. Defaults to nil if none is given.","ref":"Sourceror.html#prewalk/3","title":"Sourceror.prewalk/3","type":"function"},{"doc":"A wrapper around Code.quoted_to_algebra/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#quoted_to_algebra/2","title":"Sourceror.quoted_to_algebra/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted/2","title":"Sourceror.string_to_quoted/2","type":"macro"},{"doc":"A wrapper around Code.string_to_quoted_with_comments!/2 for compatibility with pre 1.13 Elixir versions.","ref":"Sourceror.html#string_to_quoted!/2","title":"Sourceror.string_to_quoted!/2","type":"macro"},{"doc":"Converts a quoted expression to a string. The comments line number will be ignored and the line number of the associated node will be used when formatting the code. Options :line_length - The max line length for the formatted code. :indent - how many indentations to insert at the start of each line. Note that this only prepends the indents without checking the indentation of nested blocks. Defaults to 0 . :indent_type - the type of indentation to use. It can be one of :spaces , :single_space or :tabs . Defaults to :spaces .","ref":"Sourceror.html#to_string/2","title":"Sourceror.to_string/2","type":"function"},{"doc":"Updates the arguments for the given node. iex&gt; node = { :foo , [ line : 1 ] , [ { :__block__ , [ line : 1 ] , [ 2 ] } ] } iex&gt; updater = fn args -&gt; Enum . map ( args , &amp; Sourceror . correct_lines ( &amp;1 , 2 ) ) end iex&gt; Sourceror . update_args ( node , updater ) { :foo , [ line : 1 ] , [ { :__block__ , [ line : 3 ] , [ 2 ] } ] }","ref":"Sourceror.html#update_args/2","title":"Sourceror.update_args/2","type":"function"},{"doc":"","ref":"Sourceror.html#t:comment/0","title":"Sourceror.comment/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:position/0","title":"Sourceror.position/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:range/0","title":"Sourceror.range/0","type":"type"},{"doc":"","ref":"Sourceror.html#t:traversal_function/0","title":"Sourceror.traversal_function/0","type":"type"},{"doc":"Utilities to merge an un-merge comments and quoted expressions.","ref":"Sourceror.Comments.html","title":"Sourceror.Comments","type":"module"},{"doc":"Does the opposite of merge_comments/2 , it extracts the comments from the quoted expression and returns both as a {quoted, comments} tuple.","ref":"Sourceror.Comments.html#extract_comments/2","title":"Sourceror.Comments.extract_comments/2","type":"function"},{"doc":"Merges the comments into the given quoted expression. The comments are inserted into the metadata of their closest node. Comments in the same line of before a node are inserted into the :leading_comments field while comments that are right before an end keyword are inserted into the :trailing_comments field.","ref":"Sourceror.Comments.html#merge_comments/2","title":"Sourceror.Comments.merge_comments/2","type":"function"},{"doc":"The state struct for Sourceror traversal functions.","ref":"Sourceror.TraversalState.html","title":"Sourceror.TraversalState","type":"module"},{"doc":"","ref":"Sourceror.TraversalState.html#t:t/0","title":"Sourceror.TraversalState.t/0","type":"type"},{"doc":"Implements a Zipper for the Elixir AST based on Gérard Huet Functional pearl: the zipper paper.","ref":"Sourceror.Zipper.html","title":"Sourceror.Zipper","type":"module"},{"doc":"Inserts the item as the rightmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#append_child/2","title":"Sourceror.Zipper.append_child/2","type":"function"},{"doc":"Returns true if the node is a branch.","ref":"Sourceror.Zipper.html#branch?/1","title":"Sourceror.Zipper.branch?/1","type":"function"},{"doc":"Returns a list of children of the node.","ref":"Sourceror.Zipper.html#children/1","title":"Sourceror.Zipper.children/1","type":"function"},{"doc":"Returns the zipper of the leftmost child of the node at this zipper, or nil if no there's no children.","ref":"Sourceror.Zipper.html#down/1","title":"Sourceror.Zipper.down/1","type":"function"},{"doc":"Returns true if the zipper represents the end of a depth-first walk.","ref":"Sourceror.Zipper.html#end?/1","title":"Sourceror.Zipper.end?/1","type":"function"},{"doc":"Returns a zipper to the node that satisfies the predicate function, or nil if none is found.","ref":"Sourceror.Zipper.html#find/2","title":"Sourceror.Zipper.find/2","type":"function"},{"doc":"Inserts the item as the leftmost child of the node at this zipper, without moving.","ref":"Sourceror.Zipper.html#insert_child/2","title":"Sourceror.Zipper.insert_child/2","type":"function"},{"doc":"Inserts the item as the left sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_left/2","title":"Sourceror.Zipper.insert_left/2","type":"function"},{"doc":"Inserts the item as the right sibling of the node at this zipper, without moving. Raises an ArgumentError when attempting to insert a sibling at the top level.","ref":"Sourceror.Zipper.html#insert_right/2","title":"Sourceror.Zipper.insert_right/2","type":"function"},{"doc":"Returns the zipper of the left sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#left/1","title":"Sourceror.Zipper.left/1","type":"function"},{"doc":"Returns the leftmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#leftmost/1","title":"Sourceror.Zipper.leftmost/1","type":"function"},{"doc":"Returns a new branch node, given an existing node and new children.","ref":"Sourceror.Zipper.html#make_node/2","title":"Sourceror.Zipper.make_node/2","type":"function"},{"doc":"Returns the following zipper in depth-first pre-order. When reaching the end, returns a distinguished zipper detectable via end?/1 . If it's already at the end, it stays there.","ref":"Sourceror.Zipper.html#next/1","title":"Sourceror.Zipper.next/1","type":"function"},{"doc":"Returns the node at the zipper.","ref":"Sourceror.Zipper.html#node/1","title":"Sourceror.Zipper.node/1","type":"function"},{"doc":"Returns the previous zipper in depth-first pre-order. If it's already at the end, it returns nil.","ref":"Sourceror.Zipper.html#prev/1","title":"Sourceror.Zipper.prev/1","type":"function"},{"doc":"Removes the node at the zipper, returning the zipper that would have preceded it in a depth-first walk.","ref":"Sourceror.Zipper.html#remove/1","title":"Sourceror.Zipper.remove/1","type":"function"},{"doc":"Replaces the current node in the zipper with a new node.","ref":"Sourceror.Zipper.html#replace/2","title":"Sourceror.Zipper.replace/2","type":"function"},{"doc":"&quot;Returns the zipper of the right sibling of the node at this zipper, or nil.","ref":"Sourceror.Zipper.html#right/1","title":"Sourceror.Zipper.right/1","type":"function"},{"doc":"Returns the rightmost sibling of the node at this zipper, or itself.","ref":"Sourceror.Zipper.html#rightmost/1","title":"Sourceror.Zipper.rightmost/1","type":"function"},{"doc":"Walks the zipper all the way up and returns the root node.","ref":"Sourceror.Zipper.html#root/1","title":"Sourceror.Zipper.root/1","type":"function"},{"doc":"Walks the zipper all the way up and returns the top zipper.","ref":"Sourceror.Zipper.html#top/1","title":"Sourceror.Zipper.top/1","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node. Returns a zipper to the root node. The function must return a zipper.","ref":"Sourceror.Zipper.html#traverse/2","title":"Sourceror.Zipper.traverse/2","type":"function"},{"doc":"Traverses the tree in depth-first pre-order calling the given function for each node with an accumulator. Returns a zipper to the root node and the accumulator.","ref":"Sourceror.Zipper.html#traverse/3","title":"Sourceror.Zipper.traverse/3","type":"function"},{"doc":"Returns the zipper of the parent of the node at this zipper, or nil if at the top.","ref":"Sourceror.Zipper.html#up/1","title":"Sourceror.Zipper.up/1","type":"function"},{"doc":"Replaces the current node in the zipper with the result of applying fun to the node.","ref":"Sourceror.Zipper.html#update/2","title":"Sourceror.Zipper.update/2","type":"function"},{"doc":"Creates a zipper from a tree node.","ref":"Sourceror.Zipper.html#zip/1","title":"Sourceror.Zipper.zip/1","type":"function"},{"doc":"","ref":"Sourceror.Zipper.html#t:path/0","title":"Sourceror.Zipper.path/0","type":"type"},{"doc":"","ref":"Sourceror.Zipper.html#t:tree/0","title":"Sourceror.Zipper.tree/0","type":"type"},{"doc":"","ref":"Sourceror.Zipper.html#t:zipper/0","title":"Sourceror.Zipper.zipper/0","type":"type"},{"doc":"Sourceror Utilities to work with Elixir source code. NOTICE: This library is under heavy development. Expect frequent breaking changes until the first stable v1.0 release is out.","ref":"readme.html","title":"Sourceror","type":"extras"},{"doc":"You can find Sourceror documentation on Hex Docs .","ref":"readme.html#documentation","title":"Sourceror - Documentation","type":"extras"},{"doc":"You can find usage examples in the examples folder. You can run them by cloning the repo and running elixir examples/&lt;example_file&gt;.exs . You can also find documented examples you can run with Livebook in the notebooks folder.","ref":"readme.html#examples","title":"Sourceror - Examples","type":"extras"},{"doc":"If you want to contribute to Sourceror, please check our Contributing section for pointers.","ref":"readme.html#contributing","title":"Sourceror - Contributing","type":"extras"},{"doc":"If you have any questions about Sourceror or need assistance, please open a thread in the Discussions section.","ref":"readme.html#getting-assistance","title":"Sourceror - Getting assistance","type":"extras"},{"doc":"Add :sourceror as a dependency to your project's mix.exs : defp deps do [ { :sourceror , &quot;~&gt; 0.7.0&quot; } ] end A note on compatibility Sourceror is compatible with Elixir versions down to 1.10 and OTP 21. For Elixir versions prior to 1.13 it uses a vendored version of the Elixir parser and formatter modules. This means that for Elixir versions prior to 1.12 it will successfully parse the new syntax for stepped ranges instead of raising a SyntaxError , but everything else should work as expected.","ref":"readme.html#installation","title":"Sourceror - Installation","type":"extras"},{"doc":"Be as close as possible to the standard Elixir AST. Make working with comments as simple as possible. No runtime dependencies, to simplify integration with other tools.","ref":"readme.html#goals-of-the-library","title":"Sourceror - Goals of the library","type":"extras"},{"doc":"There have been several attempts at source code manipulation in the Elixir community. Thanks to its metaprogramming features, Elixir provides builtin tools that let us get the AST of any Elixir code, but when it comes to turning the AST back to code as text, we had limited options. Macro.to_string/2 is a thing, but the produced code is generally ugly, mostly because of the extra parenthesis or because it turns string interpolations into calls to erlang modules, to name some examples. This meant that, even if we could use Macro.to_string/2 to get a string and then give that to the Elixir formatter Code.format_string!/2 , the output would still be suboptimal, as the formatter is not designed to change the semantics of the code, only to pretty print it. For example, call to erlang modules would be kept as is instead of being turned back to interpolations. We also had the additional problem of comments being discarded by the tokenizer, and literals not having information like line numbers or delimiter characters. This makes the regular AST too lossy to be useful if what we want is to manipulate the source code, because we need as much information as possible to be able to stay as close to the source as possible. There have been several proposal in the past to bring all this information to the Elixir AST, but they all meant a change that would either break macros due to the addition of new types of AST nodes, or making a compromise in core Elixir itself by storing comments in the nods metadata. Despite of all these issues, the Elixir formatter is still capable of manipulating the source code to pretty print it. Under the hood it does some neat tricks to have all this information available: on one hand, it tells the tokenizer to extract the comments from the source code and keep it at hand(not in the AST itself, but as a separate data structure), and on the other hand it tells the parser to wrap literals in block nodes so metadata can be preserved. Once it has all it needs, it can start converting the AST and comments into an algebra document, and ultimately convert that to a string. This functionality was private, and if we wanted to do it ourselves we would have to replicate or vendor the Elixir formatter with all its more than 2000 lines of code. This approach was explored by Wojtek Mach in wojtekmach/fix , but it involved vendoring the elixir Formatter code, was tightly coupled to the formatting process, and any change in Elixir would break the code. Since Elixir 1.13 this functionality from the formatter was finally exposed via the Code.string_to_quoted_with_comments/2 and Code.quoted_to_algebra/2 functions. The former gives us access to the list of comments in a shape the Elixir formatter is able to use, and the latter lets us turn any arbitrary Elixir AST into an algebra document. If we also give it the list of comments, it will merge them together, allowing us to format AST and preserve the comments. Now all we need to care about is of manipulating the AST, and let the formatter do the rest.","ref":"readme.html#background","title":"Sourceror - Background","type":"extras"},{"doc":"Having the AST and comments as separate entities allows Elixir to expose the code formatting utilities without making any changes to it's AST, but also delegates the task of figuring out what's the most appropiate way to work with them to us. Sourceror's take is to use the node metadata to store the comments. This allows us to work with an AST that is as close to regular elixir AST as possible. It also allows you to move nodes around without worrying about leaving a comment behind and ending up with misplaced comments. Two metadata fields are added to the regular Elixir AST: :leading_comments - holds the comments directly above the node or are in the same line as it. For example: test &quot;parses leading comments&quot; do quoted = &quot;&quot;&quot; # Comment for :a :a # Also a comment for :a &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , meta , [ :a ] } = quoted assert meta [ :leading_comments ] == [ %{ line : 1 , previous_eol_count : 1 , next_eol_count : 1 , text : &quot;# Comment for :a&quot; } , %{ line : 2 , previous_eol_count : 0 , next_eol_count : 1 , text : &quot;# Also a comment for :a&quot; } , ] end :trailing_comments - holds the comments that are inside of the node, but aren't leading any children, for example: test &quot;parses trailing comments&quot; do quoted = &quot;&quot;&quot; def foo() do :ok # A trailing comment end # Not a trailing comment for :foo &quot;&quot;&quot; |&gt; Sourceror . parse_string! ( ) assert { :__block__ , block_meta , [ { :def , meta , _ } ] } = quoted assert [ %{ line : 3 , text : &quot;# A trailing comment&quot; } ] = meta [ :trailing_comments ] assert [ %{ line : 4 , text : &quot;# Not a trailing comment for :foo&quot; } ] = block_meta [ :trailing_comments ] end Note that Sourceror considers leading comments to the ones that are found in the same line as a node, and trailing coments to the ones that are found before the ending line of a node, based on the end , closing or end_of_expression line. This also makes the Sourceror AST consistent with the way the Elixir formatter works, making it easier to reason about how a given AST would be formatted.","ref":"readme.html#sourceror-s-ast","title":"Sourceror - Sourceror's AST","type":"extras"},{"doc":"Copyright (c) 2021 dorgandash@gmail.com Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"readme.html#license","title":"Sourceror - License","type":"extras"}]